/**************************************************************************/
/*									  */
/* Copyright (c) 2000 by Alexandr V.Shutko, Khabarovsk, Russia		  */
/* All rights reserved.							  */
/*                                                                        */
/* Redistribution and use in source and binary forms, with or without	  */
/* modification, are permitted provided that the following conditions	  */
/* are met:								  */
/* 1. Redistributions of source code must retain the above copyright	  */
/*    notice, this list of conditions and the following disclaimer.	  */
/* 2. Redistributions in binary form must reproduce the above copyright	  */
/*    notice, this list of conditions and the following disclaimer in 	  */
/*    the documentation and/or other materials provided with the 	  */
/*    distribution.							  */
/*                                                                        */
/* THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 	  */
/* PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS  */
/* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 	  */
/* OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT   */
/* OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 	  */
/* BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,  */
/* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE   */
/* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 	  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.			  */
/*									  */
/* This unit implement database management functions (check, init, etc)   */
/* table parameters for automatic database initialization and checks 	  */
/*									  */
/* Contributor:	Alexandr V. Shutko <AVShutko@mail.khstu.ru>		  */
/* Created:	12 September 2001					  */
/* Updated:	26 December 2001 by Regressor <AVShutko@mail.khstu.ru>	  */
/*									  */
/**************************************************************************/

#include "includes.h"

/**************************************************************************/
/* This routine will check database structure and fix it in some cases 	  */
/**************************************************************************/
BOOL check_and_fix_database(char *dbname, char *dbuser, char *dbpass, 
			    char *dbaddr, char *dbport)
{
   PGconn   *check_conn;
   PGresult *res;
   BOOL check_result[8];
   cstring dbcomm_str;

   LOG_SYS(0, ("Init: Checking DB (%s) integrity...\n", dbname));
   /* First mark all tests failed */
   for(int i=0; i<8; i++) check_result[i] = False;

   /* database template1 connect */
   char	    *pgoptions, *pgtty;

   pgoptions = NULL;	/* special options for backend server   */
   pgtty     = NULL;	/* debugging tty for the backend server */

   /* check if specified database exist & mark result */
   /* make a connection to the database */
   check_conn = PQsetdbLogin(dbaddr, dbport, pgoptions, pgtty, 
                                "template1", dbuser, dbpass);

   if (PQstatus(check_conn) == CONNECTION_BAD)
   {
      LOG_SYS(0, ("Can't check database... Connection failed.\n"));
      LOG_SYS(0, ("%s", PQerrorMessage(check_conn)));
      exit(-1);
   }
	
   PQsetNoticeProcessor(check_conn, 
   (void (*)(void *, const char *))uNoticeStub, NULL);
    
   /* exec select command on backend server */
   slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
           "SELECT COUNT(*) FROM pg_database
	    WHERE datname like '%s'", dbname);

   res = PQexec(check_conn, dbcomm_str);
   if (PQresultStatus(res) != PGRES_TUPLES_OK)
   {
      LOG_SYS(0, ("SQL Command failed...\n"));
      LOG_SYS(0, ("%s", PQresultErrorMessage(res)));
      PQclear(res);
      return (False);
   }
   
   if (PQntuples(res) > 0) 
   {
      if (atoi(PQgetvalue(res, 0, 0)) > 0) 
      {
         check_result[DB_EXIST] = True;
         PQclear(res);
      }
      else
      {
         LOG_ALARM(0, ("Database %s not found, please create it 
	              with createdb tool\n", dbname));
         PQclear(res);
         exit(-1);
      }
   }
   else 
   {
      LOG_SYS(0, ("Can't fetch information about database %s\n", dbname));
      PQclear(res);
      exit(-1);
   }

   PQfinish(check_conn);
   
   /* now it is time to connect to specified database and check all tables */
   check_conn = PQsetdbLogin(dbaddr, dbport, pgoptions, pgtty, 
                                dbname, dbuser, dbpass);

   if (PQstatus(check_conn) == CONNECTION_BAD)
   {
      LOG_SYS(0, ("Can't connect db to check tables...\n"));
      LOG_SYS(0, ("%s", PQerrorMessage(check_conn)));
      exit(-1);
   }
	
   PQsetNoticeProcessor(check_conn, 
   (void (*)(void *, const char *))uNoticeStub, NULL);
   
   /* check if mess table exist and correct */
   check_result[MESS_TBL_OK] = check_user_tbl(check_conn, 
                               MESS_TBL_NAME, MESS_TBL_FIELDS);
   /* check if lock table exist and correct */
   check_result[LOCK_TBL_OK] = check_user_tbl(check_conn, 
                               LOCK_TBL_NAME, LOCK_TBL_FIELDS);   
   /* check if user table exist and correct */
   check_result[USER_TBL_OK] = check_user_tbl(check_conn, 
                               USER_TBL_NAME, USER_TBL_FIELDS);
   /* check if deps table exist and correct */
   check_result[DEPS_TBL_OK] = check_user_tbl(check_conn, 
                               DEPS_TBL_NAME, DEPS_TBL_FIELDS);
   /* check if cont table exist and correct */
   check_result[CONT_TBL_OK] = check_user_tbl(check_conn, 
                               CONT_TBL_NAME, CONT_TBL_FIELDS);
   /* check if onln table exist and correct */
   check_result[ONLN_TBL_OK] = check_user_tbl(check_conn, 
                               ONLN_TBL_NAME, ONLN_TBL_FIELDS);
   /* check if dfrg table exist and correct */
   check_result[DFRG_TBL_OK] = check_user_tbl(check_conn, 
                               DFRG_TBL_NAME, DFRG_TBL_FIELDS);

   /* delete old corrupted tables and create new one */
   if (check_result[MESS_TBL_OK] != True) 
   {
       delete_user_tbl(check_conn, MESS_TBL_NAME);
       create_user_tbl(check_conn, MESS_TBL_CREATE);
   }

   if (check_result[LOCK_TBL_OK] != True) 
   {
       delete_user_tbl(check_conn, LOCK_TBL_NAME);
       create_user_tbl(check_conn, LOCK_TBL_CREATE);
   }

   if (check_result[USER_TBL_OK] != True) 
   {
       delete_user_tbl(check_conn, USER_TBL_NAME);
       create_user_tbl(check_conn, USER_TBL_CREATE);
       create_user_tbl(check_conn, USER_TBL_CREATE1);
   }

   if (check_result[DEPS_TBL_OK] != True) 
   {
       delete_user_tbl(check_conn, DEPS_TBL_NAME);
       create_user_tbl(check_conn, DEPS_TBL_CREATE);
   }

   if (check_result[CONT_TBL_OK] != True) 
   {
       delete_user_tbl(check_conn, CONT_TBL_NAME);
       create_user_tbl(check_conn, CONT_TBL_CREATE);
   }

   if (check_result[ONLN_TBL_OK] != True) 
   {
       delete_user_tbl(check_conn, ONLN_TBL_NAME);
       create_user_tbl(check_conn, ONLN_TBL_CREATE);
   }

   if (check_result[DFRG_TBL_OK] != True) 
   {
       delete_user_tbl(check_conn, DFRG_TBL_NAME);
       create_user_tbl(check_conn, DFRG_TBL_CREATE);
   }
    
   PQfinish(check_conn);
   
   return(True);
}


/**************************************************************************/
/* This routine will check specified table			 	  */
/**************************************************************************/
BOOL check_user_tbl(PGconn *dbconn, char *tbl_name, int fldnum)
{
   PGresult *res;
   cstring dbcomm_str;

   /* exec select command on backend server */
   slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
           "SELECT COUNT(*) FROM pg_tables 
	    WHERE tablename LIKE '%s'", tbl_name);

   res = PQexec(dbconn, dbcomm_str);
   if (PQresultStatus(res) != PGRES_TUPLES_OK)
   {
      LOG_SYS(0, ("Check table: SQL cmd failed...\n"));
      LOG_SYS(0, ("%s", PQresultErrorMessage(res)));
      PQclear(res);
      return(False);
   }

   if (PQnfields(res) != 1) 
   {
      LOG_SYS(0, ("Corrypted table structure in db: 
                   \"%s\"\n", lp_db_users()));
      exit(-1);
   }
   
   if (PQntuples(res) < 0) 
   {
     LOG_SYS(0, ("Problem fetching information about 
                  table from pg_tables...\n"));
     PQclear(res); 
     return (False);
   }
   
   if (atol(PQgetvalue(res, 0, 0)) == 0) 
   {
     LOG_SYS(0, ("Can't find table \"%s\" in database...\n", tbl_name));
     PQclear(res); 
     return (False);
   }
   
   /* uff... table exist... but we should check if it have correct number 
   of fields (i think checking field names and types is not neccessary) */

   slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
           "select * FROM %s WHERE false;", tbl_name);

   res = PQexec(dbconn, dbcomm_str);
   if (PQresultStatus(res) != PGRES_TUPLES_OK)
   {
      LOG_SYS(0, ("Check table... SQL cmd failed...\n"));
      LOG_SYS(0, ("%s", PQresultErrorMessage(res))); 

      PQclear(res);
      return(False);
   }
   
   if (PQnfields(res) != fldnum) 
   {
      PQclear(res);
      LOG_SYS(0, ("Table %s has struct problem...\n",tbl_name));
      LOG_SYS(0, ("It have %d fields instead of %d...\n", PQnfields(res), fldnum));
      return(False); 
   }

   
   return(True);
}


/**************************************************************************/
/* This routine will create specified table			 	  */
/**************************************************************************/
BOOL create_user_tbl(PGconn *dbconn, char *cr_query)
{
   PGresult *res;

   LOG_SYS(0, ("Creating new database table: "));
   res = PQexec(dbconn, cr_query);
   if (PQresultStatus(res) != PGRES_COMMAND_OK)
   {
      LOG_SYS(0, ("SQL cmd failed...\n"));
      LOG_SYS(0, ("%s", PQresultErrorMessage(res)));
      PQclear(res);
      return(False);
   }

   LOG_SYS(0, ("Success...\n"));
   return(True);
}


/**************************************************************************/
/* This routine will grant permissions on spc table to specified user	  */
/**************************************************************************/
BOOL grant_permissions(PGconn *dbconn, char *tbl_name, char *username)
{
   PGresult *res;
   cstring dbcomm_str;

   slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
           "GRANT ALL ON %s TO %s", tbl_name, username);

   res = PQexec(dbconn, dbcomm_str);
   if (PQresultStatus(res) != PGRES_COMMAND_OK)
   {
      PQclear(res);
      return(False);
   }

   return(True);

}

/**************************************************************************/
/* This routine will delete spc table					  */
/**************************************************************************/
BOOL delete_user_tbl(PGconn *dbconn, char *tblname)
{
   PGresult *res;
   cstring dbcomm_str;

   slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
           "DROP TABLE %s", tblname);

   res = PQexec(dbconn, dbcomm_str);
   if (PQresultStatus(res) != PGRES_COMMAND_OK)
   {
      PQclear(res);
      return(False);
   }

   return(True);
}
