/**************************************************************************/
/*									  */
/* Copyright (c) 2000 by Alexandr V.Shutko, Khabarovsk, Russia		  */
/* All rights reserved.							  */
/*                                                                        */
/* Redistribution and use in source and binary forms, with or without	  */
/* modification, are permitted provided that the following conditions	  */
/* are met:								  */
/* 1. Redistributions of source code must retain the above copyright	  */
/*    notice, this list of conditions and the following disclaimer.	  */
/* 2. Redistributions in binary form must reproduce the above copyright	  */
/*    notice, this list of conditions and the following disclaimer in 	  */
/*    the documentation and/or other materials provided with the 	  */
/*    distribution.							  */
/*                                                                        */
/* THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 	  */
/* PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS  */
/* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 	  */
/* OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT   */
/* OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 	  */
/* BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,  */
/* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE   */
/* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 	  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.			  */
/*									  */
/* This unit implements functions related packet confirming and tracking  */
/* non-confirmed packets				 		  */
/*									  */
/* Contributor:	Alexandr V. Shutko <AVShutko@mail.khstu.ru>		  */
/* Created:	03 November 2000					  */
/* Updated:	26 December 2000 by Regressor <AVShutko@mail.khstu.ru>	  */
/*									  */
/**************************************************************************/

#include "includes.h"

/* for packet queue... */
struct indpack_record *pk_begin = NULL,
                      *pk_last  = NULL,
                      *pk_list  = NULL;

/**************************************************************************/
/* Processing event: if we have ack event - we should remove confirmed	  */
/* packet from list, if it is db_change event - we should check if online */
/* cache data for this uin is valid.					  */
/**************************************************************************/
void event_process_event(event_pack &inpack)
{
    DEBUG(200, ("EVENT: mtype: %lu, uin: %lu, stamp: %lu, ttl: %d\n", 
	     inpack.mtype, inpack.uin_number, inpack.ack_stamp, 
	     inpack.ttl));

    switch (inpack.mtype)
    {
	case MESS_ACK:
	     DEBUG(150, ("We got message: MESS_ACK\n"));
	     remove_packet_from_list(inpack.uin_number, inpack.ack_stamp);
	     break;
	case MESS_ONLINE:
	     DEBUG(150, ("We got message: MESS_ONLINE\n"));	    
	     online_cache_delete(inpack.uin_number);
             online_cache_insert(inpack.uin_number, inpack.ttl);
	     break;
	case MESS_OFFLINE:
	     DEBUG(150, ("We got message: MESS_OFFLINE\n"));
	     online_cache_delete(inpack.uin_number);
	     break;
	case MESS_UPDATE:
	     DEBUG(150, ("We got message: MESS_UPDATE (ttl=%d)\n",inpack.ttl));
	     online_cache_update(inpack.uin_number, inpack.ttl);
	     break;
	case MESS_TIMEOUT:
	     DEBUG(150, ("We got message: MESS_TIMEOUT\n"));
	     check_ask_list();
	     online_cache_decrement();
	     break;
	     
	default: break;
    }
}

/**************************************************************************/
/* Processing event packet: send packet to network socket, add		  */
/* it to list, check its timeout & retry_nums each 1 sec and 		  */
/* resend it if need.							  */
/**************************************************************************/
void event_process_packet(indirect_pack &inpack)
{
 Packet ippack; 

 memcpy(ippack.buff, inpack.buff, inpack.sizeVal);
 ippack.sizeVal   = inpack.sizeVal;
 ippack.from_ip   = inpack.to_ip;
 ippack.from_port = inpack.to_port;
 udp_send_direct_packet(ippack);
 
 add_packet_to_list(inpack);
}

/**************************************************************************/
/* Packets list  WARNING: don't use after list creation. It can		  */
/* cause a memory leak				 			  */
/**************************************************************************/
void init_packets_list()
{
 pk_begin = NULL;
 pk_list =  NULL;
 pk_last =  NULL;
}

/**************************************************************************/
/* Send packet to user and add it to list				  */
/**************************************************************************/
void add_packet_to_list(indirect_pack &inpack)
{
  try
  {
     pk_last = new indpack_record;
  }  
  catch (...)
  {
     LOG_SYS(20, ("Can't alloc memory for packet to insert in queue...\n"));
     return;
  }

  /* this is not critical so if we can't do it -> simply warn & return */
  if (pk_last == NULL) 
  {
     LOG_SYS(20, ("Can't alloc memory for packet to insert in queue...\n"));
     return;
  }
  
  DEBUG(150,("Ins pack: to:%lu, size:%d, stamp:%lu, retry_num:%d timeout:%d\n",
        inpack.to_uin, inpack.sizeVal, inpack.ask_stamp, 
	inpack.retry_num, inpack.time_out));
  
  memcpy(pk_last->ipacket.buff, inpack.buff, sizeof(inpack.buff));
  pk_last->ipacket.sizeVal   = inpack.sizeVal;
  pk_last->ipacket.time_out  = inpack.time_out;
  pk_last->ipacket.retry_num = inpack.retry_num;
  pk_last->ipacket.to_ip     = inpack.to_ip;
  pk_last->ipacket.to_port   = inpack.to_port;
  pk_last->ipacket.ask_stamp = inpack.ask_stamp;
  pk_last->ipacket.to_uin    = inpack.to_uin;

  pk_last->ipacket.ctime_out = pk_last->ipacket.time_out;
  
  if (pk_begin == NULL)     /* list is empty */
  {
    pk_last->prev = NULL;
    pk_begin = pk_last;
    pk_last->next = NULL;
  }
  else			    /* insert before first */
  {
    pk_list  = pk_begin;
    pk_begin = pk_last;
    pk_begin->next = pk_list;
    pk_begin->next->prev = pk_begin;
    pk_begin->prev = NULL;
  }
}

/**************************************************************************/
/* Remove packet from list						  */
/**************************************************************************/
void remove_packet_from_list(unsigned long ask_uin, unsigned long ask_stamp)
{
  struct indpack_record * pk_list = NULL;
  pk_list = pk_begin;

  while(pk_list)
  {
    if ((pk_list->ipacket.to_uin == ask_uin) & 
	(pk_list->ipacket.ask_stamp == ask_stamp))
    {
      DEBUG(200,("Del pack: to:%lu, size:%d, stamp:%lu\n",
        pk_list->ipacket.to_uin, pk_list->ipacket.sizeVal, 
        pk_list->ipacket.ask_stamp));

      /* delete from middle of list */
      if ((pk_list->prev != NULL) & (pk_list->next != NULL))
      {
        pk_list->prev->next = pk_list->next;	/* relink pointers */
	pk_list->next->prev = pk_list->prev;
	delete pk_list;				/* release memory */
	return;
      }
      
      /* list contain only one record */
      if ((pk_list->prev == NULL) & (pk_list->next == NULL))
      {
        pk_begin = NULL;
	delete pk_list;
	return;
      }      
      
      /* delete record from end of the list */
      if (pk_list->next == NULL)
      {
        pk_list->prev->next = NULL;
	delete pk_list;
	return;
      }
      
      /* delete record from begin of the list */
      if (pk_list->prev == NULL)
      { 
        pk_begin = pk_list->next;
	pk_list->next->prev = NULL;
	delete pk_list;
	return;
      }
    }
    pk_list = pk_list->next;
  }    
}

/**************************************************************************/
/* Check all list, decrease timeout or retry_num on each packet		  */
/* send packet again if timeout = 0. This func called every sec.	  */
/**************************************************************************/
void check_ask_list()
{
 Packet ippack;
 pk_list = pk_begin;
 struct indpack_record * ltemp = NULL;
 
 while(pk_list)
 {
   if (((pk_list->ipacket.ctime_out) == 0) & 
       ((pk_list->ipacket.retry_num) != 0)) 
   {
       DEBUG(100, ("Timeout. Resending expired packet [stamp=%lu]...\n", 
                  pk_list->ipacket.ask_stamp));
					   
       memcpy(ippack.buff, pk_list->ipacket.buff, pk_list->ipacket.sizeVal);
       ippack.sizeVal    = pk_list->ipacket.sizeVal;
       ippack.from_ip    = pk_list->ipacket.to_ip;
       ippack.from_port  = pk_list->ipacket.to_port;
       
       udp_send_direct_packet(ippack);
       
       pk_list->ipacket.retry_num--;
       pk_list->ipacket.ctime_out = pk_list->ipacket.time_out;
   }
   
   if (((pk_list->ipacket.ctime_out) == 0) & 
       ((pk_list->ipacket.retry_num) == 0)) 
   {
       packet_expire(pk_list->ipacket);
       ltemp = pk_list->next;
       remove_packet_from_list(pk_list->ipacket.to_uin, 
                               pk_list->ipacket.ask_stamp);
       pk_list = ltemp;
       continue;
   }
   
   if (pk_list->ipacket.ctime_out != 0) {
   pk_list->ipacket.ctime_out--;}
    
   pk_list = pk_list->next;
 }
}

/**************************************************************************/
/* Called when paket is expired in list	(maybe we'll want to force user	  */
/* go offline because of that)						  */
/**************************************************************************/
void packet_expire(indirect_pack &ipack)
{
  DEBUG(10, ("Packet expired in confirm list [uin:%lu], [stamp:%lu]\n",
              ipack.to_uin, ipack.ask_stamp));
}

/**************************************************************************/
/* Event processor SIGALRM handler 					  */
/**************************************************************************/
RETSIGTYPE eventSIGALRMHandler(int param)
{
    send_timeout2cache();
    return;
}


/**************************************************************************/
/* This func generate unique unsigned long number to identify packet	  */
/* and its ask packet. We get pack sequence numbers (two, but second 	  */
/* can be zero) recipient uin number, we should reproduce same number 	  */
/* from ask packet (any version) 					  */
/**************************************************************************/
unsigned long generate_ask_number(unsigned long puin, 
	      unsigned long pseq1, unsigned long pseq2)
{
 /* I simply summ them, but may be there are better algs exists */
 /* This func can produce same numbers for different users	*/
 /* this is ok - we also check uin number in confirm procedure	*/
  return (puin + pseq2 + pseq1);
}

/**************************************************************************/
/* This func send delivery confirm message (ask) to event processor.	  */
/* Called by packet processors						  */
/**************************************************************************/
void process_ask_event(unsigned long puin, unsigned long pstamp)
{
  struct event_pack pmessage;
   
  pmessage.mtype      = MESS_ACK;
  pmessage.uin_number = puin;
  pmessage.ack_stamp  = pstamp;
   
  event_send_event(pmessage);
}
