/**************************************************************************/
/*									  */
/* Copyright (c) 2000 by Alexandr V.Shutko, Khabarovsk, Russia		  */
/* All rights reserved.							  */
/*                                                                        */
/* Redistribution and use in source and binary forms, with or without	  */
/* modification, are permitted provided that the following conditions	  */
/* are met:								  */
/* 1. Redistributions of source code must retain the above copyright	  */
/*    notice, this list of conditions and the following disclaimer.	  */
/* 2. Redistributions in binary form must reproduce the above copyright	  */
/*    notice, this list of conditions and the following disclaimer in 	  */
/*    the documentation and/or other materials provided with the 	  */
/*    distribution.							  */
/*                                                                        */
/* THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 	  */
/* PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS  */
/* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 	  */
/* OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT   */
/* OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 	  */
/* BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,  */
/* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE   */
/* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 	  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.			  */
/*									  */
/* This unit implements functions related tracking user timeouts   	  */
/*									  */
/* Contributor:	Alexandr V. Shutko <AVShutko@mail.khstu.ru>		  */
/* Created:	02 November 2000					  */
/* Updated:	26 December 2000 by Regressor <AVShutko@mail.khstu.ru>	  */
/*									  */
/**************************************************************************/

#include "includes.h"

/* for online cache */
struct online_cache_record *cache_begin = NULL,
                           *cache_last  = NULL,
                           *cache_list  = NULL,
			   *cache_prev  = NULL;

/**************************************************************************/
/* Use this carefully - on non empty queue it can cause memory leak...    */
/**************************************************************************/
void init_online_cache()
{
  cache_begin = NULL;
  cache_last  = NULL;
  cache_list  = NULL;

  new_online_cache();
}

/**************************************************************************/
/* Insert new user with uin & ttl to online cache			  */
/**************************************************************************/
void online_cache_insert(unsigned long uin, unsigned short ttl)
{
  try
  {
     cache_last = new online_cache_record;
  }
  catch(...)
  {
     LOG_SYS(10, ("Can't insert new user into cache in event-processor...\n"));
     return;
  }
  
  cache_last->uin  = uin;
  cache_last->ttl  = ttl;
  
  if (cache_begin == NULL) 		// list is empty
  {
    cache_begin = cache_last;
    cache_last->next = NULL;
  }
  else					// insert before first 
  {
    cache_list  = cache_begin;
    cache_begin = cache_last;
    cache_begin->next = cache_list;
  }
}

/**************************************************************************/
/* Delete record with specified uin from online cache			  */
/**************************************************************************/
void online_cache_delete(unsigned long uin)
{
  struct online_cache_record * ltemp = NULL;
  cache_list = cache_begin;
  cache_prev = NULL;

  while(cache_list)
  {
    if (cache_list->uin == uin)
    {
      DEBUG(200,("Del online cache user : uin: %lu\n", cache_list->uin));

      // delete from middle of list
      if ((cache_prev != NULL) & (cache_list->next != NULL))
      {
        cache_prev->next = cache_list->next;	  // relink pointer
	ltemp = cache_list;
	free(ltemp);			  	  // release memory
	return;
      }
      
      // list contain only one record
      if ((cache_prev == NULL) & (cache_list->next == NULL))
      {
        cache_begin = NULL;
	free(cache_list);
	return;
      }      
      
      // delete record from end of the list
      if ((cache_list->next == NULL) & (cache_prev != NULL))
      {
        cache_prev->next = NULL;
	free(cache_list);
	return;
      }
      // delete record from begin of the list
      if ((cache_prev == NULL) & (cache_list->next != NULL))
      { 
        cache_begin = cache_list->next;
	free(cache_list);
	return;
      }
    }
    cache_prev = cache_list;
    cache_list = cache_list->next;
  }    
}

/**************************************************************************/
/* Use this func in event processor after online database change to 	  */
/* update online cache							  */
/**************************************************************************/
void online_cache_update(unsigned long uin, unsigned short ttl)
 {
   struct online_cache_record * ltemp = NULL;
   cache_list = cache_begin;
   cache_prev = NULL;

   while(cache_list)
   {
     if (cache_list->uin == uin)
     {
       DEBUG(200,("Update online cache user : uin: %lu\n", cache_list->uin));     
       cache_list->ttl = ttl;
       return;
     }
     cache_prev = cache_list;
     cache_list = cache_list->next;
   }    
}

/**************************************************************************/
/* Decr counters in all cache records and force offline expired users     */
/**************************************************************************/
void online_cache_decrement()
{
  cache_list = cache_begin;
  struct online_user temp_user;
  unsigned long temp_uin;

  DEBUG(300, ("Decrement user's cache called...\n"));
  
  while(cache_list)
  {
    cache_list->ttl -= 1;
    
    if (cache_list->ttl < 1)
    {
      /* user expired, we should check db info to be sure */
      if (db_online_lookup(cache_list->uin, temp_user) == 0)
      {
         if (arb(time(NULL)-longToTime(temp_user.lutime)) > temp_user.ttl)
	 {
	   /* someone (may be another server) updated online database 
 	      information, so we should update cache ttl */
	   cache_list->ttl = time(NULL)-longToTime(temp_user.lutime);
	   DEBUG(100, ("Cache expired, but DB not. Time for user %lu left: %d sec\n", 
	               cache_list->uin, cache_list->ttl));
	 }
	 else
	 {
	   /* user expired ether in cache and database, 
	      we should remove it from cache and database */
	      
	   unsigned long uptime;
	   
	   temp_uin = cache_list->uin;
	   cache_list = cache_list->next;
	   
	   /* calculate aprox user uptime */
	   uptime = longToTime(temp_user.lutime) 
	            - longToTime(temp_user.uptime) 
	            + temp_user.ttl;
	 
	   LOG_USR(0, ("User %lu moved offline. Online time was: %lu seconds\n", temp_uin, uptime));
	   DEBUG(100, ("Online DB & cache expired. User %lu deleted\n", temp_uin));
	   move_user_offline(temp_uin);
	   continue;
	 }
      }
      else
      {
        /* if there is no user in database we should 
	   delete it from cache */
	   temp_uin = cache_list->uin;
	   cache_list = cache_list->next;

	   DEBUG(100, ("Cache expired. User %lu deleted...\n", temp_uin));
	   move_user_offline(temp_uin);
	   
	   continue;
      }
    }
    cache_list = cache_list->next;
  }
}

/**************************************************************************/
/* Delete all cache and make new cache from online database 		  */
/**************************************************************************/
void new_online_cache()
{
   PGresult *res;
   fstring dbcomm_str;
   struct online_user temp_user;
   
   purge_online_cache();

   /* exec select command on backend server */
   slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
           "SELECT * FROM Online_Users");

   res = PQexec(users_dbconn, dbcomm_str);
   if (PQresultStatus(res) != PGRES_TUPLES_OK)
   {
      LOG_SYS(0, ("SQL Command failed...\n"));
      LOG_SYS(0, ("%s", PQresultErrorMessage(res)));
      PQclear(res);
      return;
   }

   if (PQnfields(res) < 10) 
   {
      LOG_SYS(0, ("Corrypted table structure in db: \"%s\"\n", 
                  lp_db_users()));
      exit(-1);
   }
   
   if (PQntuples(res) > 0)
   {
      for (int i=0; i < PQntuples(res); i++)
      {
         temp_user.uin    = atol(PQgetvalue(res, i, 0));
	 temp_user.ttl    = atol(PQgetvalue(res, i, 7));
	 temp_user.lutime = atol(PQgetvalue(res, i, 6));
	 
         if ( arb(time(NULL)-(longToTime(temp_user.lutime)) ) < temp_user.ttl)
	 {
	    DEBUG(200, ("NEW: Inserting new record to cache: [uin:%lu][ttl:%ld]\n",
			temp_user.uin, temp_user.ttl));
	    
	    if (time(NULL)-(longToTime(temp_user.lutime)) > 0)
	    {
               online_cache_insert(temp_user.uin, 
	            (temp_user.ttl - (time(NULL)-(longToTime(temp_user.lutime)))));
	    }
	    else
	    {
	       online_cache_insert(temp_user.uin, temp_user.ttl);
	    }
	 }
	 else
	 {
	    DEBUG(200, ("NEW: Deleting expired record from db: [uin:%lu][ttl:%ld]\n",
			temp_user.uin, temp_user.ttl));
	    db_online_delete(temp_user.uin);
	    db_contact_delete(temp_user.uin);
	 }
      }
      PQclear(res);
   }
   else 
   {
      PQclear(res);
   }
}

/**************************************************************************/
/* Purge online cache							  */
/**************************************************************************/
void purge_online_cache()
{
  cache_list = cache_begin;

  DEBUG(50, ("Purging online cache....\n"));
  
  while(cache_list)
  {
    cache_last = cache_list->next;
    free(cache_list);
    cache_list = cache_last;
  }
  cache_begin = NULL;
}

/**************************************************************************/
/* This function used to move user online. It insert new record in	  */
/* cache and database and init lastupdate value in online user record	  */
/**************************************************************************/
void move_user_online(struct online_user &user)
{  
    db_users_touch( user );
    broadcast_online( user );
    db_defrag_delete( user.uin );
    send_online2cache( user.uin, user.ttl );
    
    DEBUG(100, ("User %lu moved online: [ttl=%ld,ltime=%s]\n", 
		user.uin, user.ttl, time2str(longToTime(user.lutime))));
}

/**************************************************************************/
/* This function used to move user offline. It delete cache record and	  */
/* wipe out it from database						  */
/**************************************************************************/
void move_user_offline(unsigned long uin)
{
    db_online_delete( uin );
    db_contact_delete( uin );
    db_defrag_delete( uin );

    broadcast_offline( uin );
    send_offline2cache( uin );
}

/**************************************************************************/
/* This function send message to online cache. It user by packet 	  */
/* processors to inform cache about database update.			  */
/**************************************************************************/
void send_offline2cache(unsigned long uin)
{
  struct event_pack pmessage;
   
  pmessage.mtype      = MESS_OFFLINE;
  pmessage.uin_number = uin;
  pmessage.ack_stamp  = 0;
  pmessage.ttl	      = 0;
   
  event_send_event(pmessage);
}


/**************************************************************************/
/* This function send message to online cache. It user by packet 	  */
/* processors to inform cache about database update.			  */
/**************************************************************************/
void send_online2cache(unsigned long uin, unsigned short ttl)
{
  struct event_pack pmessage;
   
  pmessage.mtype      = MESS_ONLINE;
  pmessage.uin_number = uin;
  pmessage.ack_stamp  = 0;
  pmessage.ttl	      = ttl;
  
  event_send_event(pmessage);
}


/**************************************************************************/
/* This function send message to online cache. It user by packet 	  */
/* processors to inform cache about database update.			  */
/**************************************************************************/
void send_update2cache(unsigned long uin, unsigned short ttl)
{
  struct event_pack pmessage;
  
  pmessage.mtype      = MESS_UPDATE;
  pmessage.uin_number = uin;
  pmessage.ack_stamp  = 0;
  pmessage.ttl	      = ttl;

  DEBUG(300, ("Trying to send update cmd to pipe: ttl=%d\n", pmessage.ttl));
  
  event_send_event(pmessage);
}
 
/**************************************************************************/
/* This function send message to online cache. It user by packet 	  */
/* processors to inform cache about database update.			  */
/**************************************************************************/
void send_timeout2cache()
{
  struct event_pack pmessage;
  
  pmessage.mtype      = MESS_TIMEOUT;
  pmessage.uin_number = 0;
  pmessage.ack_stamp  = 0;
  pmessage.ttl	      = 0;

  event_send_event(pmessage);
}
