/**************************************************************************/
/*									  */
/* Copyright (c) 2000 by Alexandr V.Shutko, Khabarovsk, Russia		  */
/* All rights reserved.							  */
/*									  */
/* Redistribution and use in source and binary forms, with or without	  */
/* modification, are permitted provided that the following conditions	  */
/* are met:								  */
/* 1. Redistributions of source code must retain the above copyright	  */
/*    notice, this list of conditions and the following disclaimer.	  */
/* 2. Redistributions in binary form must reproduce the above copyright	  */
/*    notice, this list of conditions and the following disclaimer in 	  */
/*    the documentation and/or other materials provided with the 	  */
/*    distribution.							  */
/*									  */
/* THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 	  */
/* PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS  */
/* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 	  */
/* OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT   */
/* OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 	  */
/* BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,  */
/* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE   */
/* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 	  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.			  */
/*									  */
/*									  */
/* This unit contain functions related broadcasting - status broadcasting */
/* message broadcasting, alert broadcasting. This unit work with database */
/* directly to increase perfomance and reduce database queries 		  */
/**************************************************************************/

#include "includes.h"

/**************************************************************************/
/* Purpose: online event broadcasting for users wanted this (by contact)  */
/* This will cause 3 complex database queries (instead of (csz-ofn)*3+1)  */
/**************************************************************************/
void broadcast_online( struct online_user &auser )
{
   PGresult *res;			/* struct for database requests   */
   fstring dbcomm_str;  		/* string space for query         */
   int users_cnt;			/* number of users to alert       */
   struct online_user to_user;		/* structure for target user      */

   /* first of all we should request online users that have target 	  */
   /* in their contacts and after that lock database rows          	  */

   PQclear( PQexec(users_dbconn, "BEGIN" ));
   slprintf(dbcomm_str, sizeof( dbcomm_str )-1, 
           "SELECT uin,prv,sseq,uprt,iadr,sess_id,tcpver FROM online_users 
	    WHERE uin IN (SELECT ouin FROM online_contacts 
	    WHERE tuin=%lu) FOR UPDATE", auser.uin);

   res = PQexec(users_dbconn, dbcomm_str);
   if (PQresultStatus( res ) != PGRES_TUPLES_OK )
   {
      LOG_SYS(0, ( "SQL Command [SELECT FOR UPDATE] failed...\n" ));
      LOG_SYS(0, ( "%s", PQresultErrorMessage( res )));
      PQclear( res );
      return;
   }
   
   /* number of returned users (tuples with information) */
   users_cnt = PQntuples( res );   
   DEBUG(100, ("Complex query return %d users to alert...\n", users_cnt));
   
   /* now it is time to send online alert to all users in loop */
   for( int i=0; i<users_cnt; i++ )
   {
     
     if ( PQnfields( res ) == 7 )   /* sanity check - to avoid core dumps */
     {
      /* first fill online_user structure */
      to_user.uin	 = atol( PQgetvalue(res, i, 0));
      to_user.protocol	 = (unsigned short)atoi(PQgetvalue(res, i, 1)); 
      to_user.servseq	 = atol(PQgetvalue(res, i, 2)); 
      to_user.udp_port	 = atol(PQgetvalue(res, i, 3)); 
      to_user.ip	 = icqToIp(atol(PQgetvalue(res, i, 4))); 
      to_user.session_id = atol(PQgetvalue(res, i, 5));
      to_user.tcpver     = atol(PQgetvalue(res, i, 6));
     
      /* all ready we can send packet */
      send_user_online( to_user, auser );
     }
   }
 
   PQclear( res );
   
   /* well, if we doesn't unlock database server will freeze for      	  */ 
   /* several users, also we should update sseq for all alerted users 	  */
   slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
           "UPDATE online_users SET sseq=sseq+1 WHERE uin IN 
	   (SELECT ouin FROM online_contacts WHERE tuin=%lu)", auser.uin);
   /* send query to backend */
   res = PQexec(users_dbconn, dbcomm_str);
   if (PQresultStatus( res ) != PGRES_COMMAND_OK)
   {
      LOG_SYS(0, ( "SQL Command [UPDATE SSEQ] failed...\n" ));
      LOG_SYS(0, ( "%s", PQresultErrorMessage( res )));
      PQclear( res );
      return;
   }  
   
   PQclear(res);   
   PQclear( PQexec(users_dbconn, "END" ));
}

/**************************************************************************/
/* Purpose: offline event broadcasting for users wanted this (by contact) */
/* This will cause 3 complex database queries (instead of (csz-ofn)*3+1)  */
/**************************************************************************/
void broadcast_offline( unsigned long uin )
{
   PGresult *res;			/* struct for database requests   */
   fstring dbcomm_str;  		/* string space for query         */
   int users_cnt;			/* number of users to alert       */
   struct online_user to_user;		/* structure for target user      */

   /* first of all we should request online users that have target 	  */
   /* in their contacts and after that lock database rows          	  */
   
   PQclear( PQexec(users_dbconn, "BEGIN" ));
   slprintf(dbcomm_str, sizeof( dbcomm_str )-1, 
           "SELECT uin,prv,sseq,uprt,iadr,sess_id FROM online_users 
	    WHERE uin IN (SELECT ouin FROM online_contacts 
	    WHERE tuin=%lu) FOR UPDATE", uin);
	            
   res = PQexec(users_dbconn, dbcomm_str);
   if (PQresultStatus( res ) != PGRES_TUPLES_OK )
   {
      LOG_SYS(0, ( "SQL Command [SELECT FOR UPDATE] failed...\n" ));
      LOG_SYS(0, ( "%s", PQresultErrorMessage( res )));
      PQclear( res );
      return;
   }
   
   /* number of returned users (tuples with information) */
   users_cnt = PQntuples( res );   
   /* now it is time to send online alert to all users in loop */
   for( int i=0; i<users_cnt; i++ )
   {
     
     if ( PQnfields( res ) == 6 )   /* sanity check - to avoid core dumps */
     {
      /* first fill online_user structure */
      to_user.uin	 = atol( PQgetvalue(res, i, 0));
      to_user.protocol	 = (unsigned short)atoi(PQgetvalue(res, i, 1)); 
      to_user.servseq	 = atol(PQgetvalue(res, i, 2)); 
      to_user.udp_port	 = atol(PQgetvalue(res, i, 3)); 
      to_user.ip	 = icqToIp(atol(PQgetvalue(res, i, 4))); 
      to_user.session_id = atol(PQgetvalue(res, i, 5));
     
      /* all ready we can send packet */
      send_user_offline( to_user, uin );
     }
   }
 
   PQclear( res );
 
   /* well, if we doesn't unlock database server will freeze for      	  */ 
   /* several users, also we should update sseq for all alerted users 	  */
   slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
           "UPDATE online_users SET sseq=sseq+1 WHERE uin IN 
	   (SELECT ouin FROM online_contacts WHERE tuin=%lu)", uin);
   /* send query to backend */
   res = PQexec(users_dbconn, dbcomm_str);
   if (PQresultStatus( res ) != PGRES_COMMAND_OK)
   {
      LOG_SYS(0, ( "SQL Command [UPDATE SSEQ] failed...\n" ));
      LOG_SYS(0, ( "%s", PQresultErrorMessage( res )));
      PQclear( res );
      return;
   }  

   PQclear(res);   
   PQclear( PQexec(users_dbconn, "END" ));
}


/**************************************************************************/
/* Purpose: Status change event broadcasting				  */
/* This will cause 3 complex database queries (instead of (csz-ofn)*3+1)  */
/**************************************************************************/
void broadcast_status( struct online_user &auser )
{
   PGresult *res;			/* struct for database requests   */
   fstring dbcomm_str;  		/* string space for query         */
   int users_cnt;			/* number of users to alert       */
   struct online_user to_user;		/* structure for target user      */

   /* first of all we should request online users that have target 	  */
   /* in their contacts and after that lock database rows          	  */

   PQclear( PQexec(users_dbconn, "BEGIN" ));
   slprintf(dbcomm_str, sizeof( dbcomm_str )-1, 
           "SELECT uin,prv,sseq,uprt,iadr,sess_id FROM online_users 
	    WHERE uin IN (SELECT ouin FROM online_contacts 
	    WHERE tuin=%lu) FOR UPDATE", auser.uin);
	            
   res = PQexec(users_dbconn, dbcomm_str);
   if (PQresultStatus( res ) != PGRES_TUPLES_OK )
   {
      LOG_SYS(0, ( "SQL Command [SELECT FOR UPDATE] failed...\n" ));
      LOG_SYS(0, ( "%s", PQresultErrorMessage( res )));
      PQclear( res );
      return;
   }
   
   /* number of returned users (tuples with information) */
   users_cnt = PQntuples( res );   
   /* now it is time to send online alert to all users in loop */
   for( int i=0; i<users_cnt; i++ )
   {
     
     if ( PQnfields( res ) == 6 )   /* sanity check - to avoid core dumps */
     {
      /* first fill online_user structure */
      to_user.uin	 = atol( PQgetvalue(res, i, 0));
      to_user.protocol	 = (unsigned short)atoi(PQgetvalue(res, i, 1)); 
      to_user.servseq	 = atol(PQgetvalue(res, i, 2)); 
      to_user.udp_port	 = atol(PQgetvalue(res, i, 3)); 
      to_user.ip	 = icqToIp(atol(PQgetvalue(res, i, 4))); 
      to_user.session_id = atol(PQgetvalue(res, i, 5));
     
      /* all ready we can send packet */
      send_user_status( to_user, auser );
     }
   }
 
   PQclear( res );
 
   /* well, if we doesn't unlock database server will freeze for      	  */ 
   /* several users, also we should update sseq for all alerted users 	  */
   slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
           "UPDATE online_users SET sseq=sseq+1 WHERE uin IN 
	   (SELECT ouin FROM online_contacts WHERE tuin=%lu)", auser.uin);
   /* send query to backend */
   res = PQexec(users_dbconn, dbcomm_str);
   if (PQresultStatus( res ) != PGRES_COMMAND_OK)
   {
      LOG_SYS(0, ( "SQL Command [UPDATE SSEQ] failed...\n" ));
      LOG_SYS(0, ( "%s", PQresultErrorMessage( res )));
      PQclear( res );
      return;
   }  

   PQclear(res);   
   PQclear( PQexec(users_dbconn, "END" ));
}

/**************************************************************************/
/* Purpose: Broadcast administrator message to online (and possibly to    */
/* offline) users.							  */
/**************************************************************************/
void broadcast_message(BOOL online_only, unsigned long department, 
		       char *message, unsigned short mess_type, 
		       unsigned long from_uin)
{
   PGresult *res;		/* structure pointer for db requests */
   cstring dbcomm_str;		/* string for db queries */
   long users_cnt;
   struct online_user to_user;
   
   /* ------------------------------------------ */
   /* First stage: Send messages to online users */
   /* ------------------------------------------ */
   
   PQclear( PQexec(users_dbconn, "BEGIN" )); /* start transaction */
   if (department != 0) 	/* if depart = 0 we'll send msg 2 all */
   {
      slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
              "SELECT uin,prv,sseq,uprt,iadr,sess_id FROM Online_Users WHERE 
	       uin IN (SELECT uin FROM users_info WHERE wdepart='%lu') 
	       and uin<>%lu FOR UPDATE", department, from_uin);
   }
   else 
   {
      slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
              "SELECT uin,prv,sseq,uprt,iadr,sess_id FROM Online_Users 
	      WHERE uin<>%lu FOR UPDATE", from_uin);
   }
   
   res = PQexec(users_dbconn, dbcomm_str);
   if (PQresultStatus(res) != PGRES_TUPLES_OK)
   {
      /* something goes wrong */
      LOG_SYS(0, ("BCAST ONLINE SELECT SQL query failed...\n"));
      LOG_SYS(0, ("%s", PQresultErrorMessage(res)));
      PQclear(res);
      return;
   }
   
   users_cnt = PQntuples(res);
   
     /* now it is time to send online alert to all users in loop */
   for( int i=0; i<users_cnt; i++ )
   {
      /* first fill online_user structure */
      to_user.uin	 = atol( PQgetvalue(res, i, 0));
      to_user.protocol	 = (unsigned short)atoi(PQgetvalue(res, i, 1)); 
      to_user.servseq	 = atol(PQgetvalue(res, i, 2)); 
      to_user.udp_port	 = atol(PQgetvalue(res, i, 3)); 
      to_user.ip	 = icqToIp(atol(PQgetvalue(res, i, 4))); 
      to_user.session_id = atol(PQgetvalue(res, i, 5));
     
      /* all things ready, now we can send packet */
      send_online_message(from_uin, to_user, mess_type, 0, message);
   }

   PQclear(res);    
   
   if (department != 0) 	/* if depart = 0 we'll send msg 2 all */
   {
      slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
              "UPDATE Online_Users SET sseq=sseq+1 WHERE uin IN 
	       (SELECT uin FROM users_info WHERE wdepart='%lu') 
	       AND uin<>%lu", department, from_uin);
   }
   else
   {
      slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
              "UPDATE Online_Users SET sseq=sseq+1 WHERE uin<>%lu", 
	       from_uin);
   }
  
   res = PQexec(users_dbconn, dbcomm_str);
   if (PQresultStatus(res) != PGRES_COMMAND_OK)
   {
      /* something goes wrong */
      LOG_SYS(0, ("BCAST ONLINE UPDATE SQL query failed...\n"));
      LOG_SYS(0, ("%s", PQresultErrorMessage(res)));
      PQclear(res);
      return;
   }
   /* we don't want memory leaks :)) */
   PQclear(res);
   PQclear( PQexec(users_dbconn, "END" ));
   
   /* -------------------------------------------- */
   /* Second stage: Send messages to offline users */
   /* -------------------------------------------- */

   PQclear( PQexec(users_dbconn, "BEGIN" ));   
   if (online_only != True)    
   {
      if (department != 0) 	/* if depart = 0 we'll send msg 2 all */
      {
         /* we should get uins for all unconnected */
	 /* users from specified department */
         slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
                 "SELECT uin FROM Users_Info WHERE uin NOT IN 
	          (SELECT uin FROM Online_Users) AND wdepart='%lu'", 
	          department);
      }
      else 
      {
         /* we should get uins for all unconnected users */
         slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
                 "SELECT uin FROM Users_Info WHERE uin NOT IN
	         (SELECT uin FROM Online_Users)");
      }
      
      res = PQexec(users_dbconn, dbcomm_str);
      if (PQresultStatus(res) != PGRES_TUPLES_OK)
      {
         /* something goes wrong */
         LOG_SYS(0, ("BCAST OFFLINE SELECT SQL query failed...\n"));
         LOG_SYS(0, ("%s", PQresultErrorMessage(res)));
         PQclear(res);
         return;
      }

      users_cnt = PQntuples(res); /* number of found users */
      
      PGresult *res1;		  /* second struct for queries */
      bigstring nmessage;	  /* for converted message */
      time_t stime, mtime;	  /* for time values */
      bigstring dbcomm_str;	  /* for query body */
      
      /* translate to server and remove all system chars */
      convert_to_postgres(nmessage, sizeof(nmessage)-1 ,message);
      /* prepare long time structure with ajusted timezone */
      stime = time(NULL); mtime = mktime(gmtime(&stime));
      
      /* now it is time to add offline message to all users in loop */
      for( long i=0; i<users_cnt; i++ )
      {
        /* get target user uin from query result */
        to_user.uin = atol( PQgetvalue(res, i, 0));
        /* prepare query string with message */
        slprintf(dbcomm_str, sizeof(dbcomm_str)-1, 
                "INSERT INTO Users_Messages values (%lu, %lu, %lu, %d, '%s')", 
                 to_user.uin, from_uin, mtime, mess_type, nmessage);
        /* exec select command on backend server */
        res1 = PQexec(users_dbconn, dbcomm_str);
        if (PQresultStatus(res1) != PGRES_COMMAND_OK)
        {
	    /*  we have problem - sql failed */
            LOG_SYS(0, ("SQL Command failed...\n"));
            LOG_SYS(0, ("%s", PQresultErrorMessage(res1)));
            PQclear(res1);
            return;
        }
	PQclear(res1);	/* to avoid memory leaks */
      }      
      PQclear( res );   /* to avoid memory leaks */
   }
   
   /* now it is time to finish transaction */
   PQclear( PQexec(users_dbconn, "END" ));
}
