/**************************************************************************/
/*									  */
/* Copyright (c) 2000 by Alexandr V.Shutko, Khabarovsk, Russia		  */
/* All rights reserved.							  */
/*									  */
/* Redistribution and use in source and binary forms, with or without	  */
/* modification, are permitted provided that the following conditions	  */
/* are met:								  */
/* 1. Redistributions of source code must retain the above copyright	  */
/*    notice, this list of conditions and the following disclaimer.	  */
/* 2. Redistributions in binary form must reproduce the above copyright	  */
/*    notice, this list of conditions and the following disclaimer in 	  */
/*    the documentation and/or other materials provided with the 	  */
/*    distribution.							  */
/*									  */
/* THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 	  */
/* PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS  */
/* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 	  */
/* OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT   */
/* OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 	  */
/* BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,  */
/* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE   */
/* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 	  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.			  */
/*									  */
/**************************************************************************/

#include "includes.h"

/**************************************************************************/
/* This func called by packet_processors. It make protocol switching	  */
/* between specific protocol handlers, if protocol is not supported it 	  */
/* dump to log notice message with level 10 				  */
/**************************************************************************/
void handle_icq_packet(Packet &pack)
{
 short unsigned int prot_vers;
 short unsigned int command_num;
 
 /* get some data from packet and reset it to begin  */
 pack >> prot_vers;	/* packet version	     */
 pack >> command_num;	/* command code		     */
 pack.reset();		/* set data pointer to begin */

 /* now we should try to determine packet proto version. When you 
 adding new proto - you should add case string bellow with proper 
 proto handler. Rules to make handlers are in docum/new_proto.txt */

 switch (prot_vers)
 {
   case V2_PROTO:  NO_PROTO_STUB(V2_PROTO);
   case V3_PROTO:  handle_v3_proto(pack);
                   break;
   case V4_PROTO:  NO_PROTO_STUB(V4_PROTO);
   case V5_PROTO:  handle_v5_proto(pack);
   case SYS_PROTO: break;
   
   default: 	  NO_PROTO_STUB(prot_vers);

 }
}

/**************************************************************************/
/* This func make initialization of all available protocols. If you are	  */
/* developing new proto - insert your init funct here.			  */
/**************************************************************************/
void init_all_protocols()
{
  v3_proto_init();   /* Initialization of V3 proto variables */
  v5_proto_init();   /* Initialization of V5 proto variables */
  
}

/**************************************************************************/
/* This function select proper function to send user_online packet 	  */
/**************************************************************************/
void send_user_online(struct online_user &to_user, struct online_user &user)
{
    DEBUG(100, ( "Sending user's %lu online status to %lu\n", 
		user.uin, to_user.uin ));
		
    switch (to_user.protocol)
    {
      case V3_PROTO: v3_send_user_online( to_user, user ); break;
      case V4_PROTO: return;
      case V5_PROTO: v5_send_user_online( to_user, user ); break;
      default: return;
    }
}

/**************************************************************************/
/* This function select proper function to send user_ch_status packet	  */
/**************************************************************************/
void send_user_status(struct online_user &to_user, struct online_user &user)
{
    DEBUG(100, ("Sending user's %lu new status to %lu\n", 
		user.uin, to_user.uin));
    
    switch (to_user.protocol)
    {
      case V3_PROTO: v3_send_user_status(to_user, user); break;
      case V4_PROTO: return;
      case V5_PROTO: v5_send_user_status(to_user, user); break;
      default: return;
    }
}


/************************************************************************/
/* This function select proper function to send user_offline packet 	*/
/************************************************************************/
void send_user_offline(struct online_user &to_user, unsigned long uin)
{
    DEBUG(100, ("Sending user's %lu offline to %lu\n", 
		uin, to_user.uin));
    
    switch (to_user.protocol)
    {
      case V3_PROTO: v3_send_user_offline(to_user, uin); break;
      case V4_PROTO: return;
      case V5_PROTO: v5_send_user_offline(to_user, uin); break;
      default: return;
    }
}


/**************************************************************************/
/* This func send protocol independent offline message to user		  */
/**************************************************************************/
void send_online_message(struct online_user &user, 
			 struct online_user &to_user, 
			 int type, unsigned short seq2,
			 char *message)
{
    int pack_num = 0;
    
    DEBUG(50, ("Sending system message (type: %d) to user %lu from %lu\n", 
		type, to_user.uin, user.uin));
    
    db_online_sseq_open(to_user);
    
    switch (to_user.protocol)
    {
      case V3_PROTO: pack_num = v3_send_user_message(user.uin, to_user, 
                                type, seq2, message); break;
      case V4_PROTO: return;
      case V5_PROTO: return;
      default: return;
    }
    
    db_online_sseq_close(to_user, pack_num);
}

/**************************************************************************/
/* This func send protocol independent offline message to user	      	  */
/* with out updating db sequence (for broadcast messages )	      	  */
/**************************************************************************/
void send_online_message(unsigned long user_uin, 
			 struct online_user &to_user, 
			 int type, unsigned short seq2,
			 char *message)
{
    DEBUG(50, ("Sending system message (type: %d) to user %lu from %lu\n", 
		type, to_user.uin, user_uin));
    
    switch (to_user.protocol)
    {
      case V3_PROTO: v3_send_user_message(user_uin, to_user, type, seq2, message); break;
      case V4_PROTO: return;
      case V5_PROTO: return;
      default: return;
    }
    
}

/**************************************************************************/
/* This func return max user timeout for specified protocol version	  */
/**************************************************************************/
int get_ping_time(int protocol)
{
  switch (protocol)
  {
    case V3_PROTO: return(lp_v3_pingtime());
    case V4_PROTO: return(lp_default_ping());
    case V5_PROTO: return(lp_v5_pingtime());
    default: return(lp_default_ping()); break;
  }
}
