/**************************************************************************/
/*									  */
/* Copyright (c) 2000 by Alexandr V.Shutko, Khabarovsk, Russia		  */
/* All rights reserved.							  */
/*									  */
/* Redistribution and use in source and binary forms, with or without	  */
/* modification, are permitted provided that the following conditions	  */
/* are met:								  */
/* 1. Redistributions of source code must retain the above copyright	  */
/*    notice, this list of conditions and the following disclaimer.	  */
/* 2. Redistributions in binary form must reproduce the above copyright	  */
/*    notice, this list of conditions and the following disclaimer in 	  */
/*    the documentation and/or other materials provided with the 	  */
/*    distribution.							  */
/*									  */
/* THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 	  */
/* PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS  */
/* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 	  */
/* OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT   */
/* OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 	  */
/* BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,  */
/* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE   */
/* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 	  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.			  */
/*									  */
/* Description: Interprocess (IPC) pipes via unix domain sockets 	  */
/* There was too many problems with pipes, so i decided to move to unix   */
/* domain sockets by changing pipe_receive_packet and pipe_send_packet    */
/**************************************************************************/

#include "includes.h"

/**************************************************************************/
/* Event pipe initialization. It return socket descriptor 	 	  */
/**************************************************************************/
void event_pipe_init()
{
  if ((event_pipe_fd = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0)
  {
     LOG_SYS(0,("ERROR: Can't create unix domain socket for IPC pipe...\n"));
     exit(-1);
  }
  
  bzero(&ep_serv_addr, sizeof(ep_serv_addr));
  ep_serv_addr.sun_family = AF_UNIX;
  snprintf(ep_serv_addr.sun_path, sizeof(ep_serv_addr.sun_path)-1,
           "%s/pipe_evt.%u", lp_var_path(), lp_udp_port());
	   
  DEBUG(90, ("Binding pipe socket to: %s\n", ep_serv_addr.sun_path));
  
  ep_saddrlen = sizeof(ep_serv_addr.sun_family) + 
                strlen(ep_serv_addr.sun_path)+1;

  if (unlink(ep_serv_addr.sun_path) == 0)
  {
     LOG_SYS(0, ("Init: Unix socket (event) exist: unclean shutdown?...\n"));
  }

  if (bind(event_pipe_fd, (struct sockaddr *)&ep_serv_addr, ep_saddrlen) < 0)
  {
     LOG_SYS(0, ("ERROR: Can't bind pipe socket to special file...\n"));
     exit(-1);
  }
  
  chmod(ep_serv_addr.sun_path, 0600);
}

/**************************************************************************/
/* Write event record to event pipe.					  */
/**************************************************************************/
void event_send_event(event_pack &pack) 
{
   /* preparing temp packet for pipe socket */
   pipe_pack.clearPacket();
   
   pipe_pack << (unsigned  long)pack.mtype
             << (unsigned  long)pack.uin_number
	     << (unsigned  long)pack.ack_stamp
	     << (unsigned short)pack.ttl;
   
   /* send packet to pipe socket */
   sendto(event_pipe_fd, pipe_pack.buff, pipe_pack.sizeVal, 0, 
          (struct sockaddr *)&ep_serv_addr, ep_saddrlen);

}

/**************************************************************************/
/* this used to receive event notifications from event pipe		  */
/**************************************************************************/
void event_receive_event(event_pack &inpack)
{
   pipe_pack.sizeVal = recvfrom(event_pipe_fd, pipe_pack.buff, 
            MAX_PACKET_SIZE, 0, (struct sockaddr *)&ep_serv_addr, 
	    &ep_saddrlen);

   DEBUG(400, ("Received event record, length: %d\n", pipe_pack.sizeVal));
   
   if (pipe_pack.sizeVal > 0)
   {
     memcpy(&inpack.mtype,      pipe_pack.buff, sizeof(unsigned long));
     memcpy(&inpack.uin_number, pipe_pack.buff+ sizeof(unsigned long),     sizeof(unsigned long));
     memcpy(&inpack.ack_stamp,  pipe_pack.buff+ (sizeof(unsigned long)*2), sizeof(unsigned long));
     memcpy(&inpack.ttl,  	pipe_pack.buff+ (sizeof(unsigned long)*3), sizeof(unsigned short));
   }
   return;
}

/**************************************************************************/
/* Write size of packet to socket domain pipe, then write 		  */
/* packet,ip and port							  */
/**************************************************************************/
void pipe_send_packet(Packet &pack)
{
   lock_incw();

   /* preparing temp packet for pipe socket */
   memcpy(pipe_pack.buff, pack.buff, pack.sizeVal);
   pipe_pack.sizeVal = pack.sizeVal;
   pipe_pack.append();
   
   pipe_pack << (unsigned long)(ipToIcq(pack.from_ip))
             << (unsigned long)(pack.from_port);

   /* send packet to pipe socket */
   sendto(incoming_pipe_fd, pipe_pack.buff, pipe_pack.sizeVal, 0, 
          (struct sockaddr *)&ip_serv_addr, ip_saddrlen);

   unlock_incw();
}

/**************************************************************************/
/* Pipe packet read procedure. Lock pipe and read packet		  */ 
/* into packet class							  */
/**************************************************************************/
void pipe_receive_packet(Packet &pack)
{
   pack.clearPacket();
   pipe_pack.clearPacket();
   lock_pipe(INCOMING_PIPE);

   pipe_pack.sizeVal = recvfrom(incoming_pipe_fd, pipe_pack.buff, 
            MAX_PACKET_SIZE, 0, (struct sockaddr *)&ip_serv_addr, 
	    &ip_saddrlen);

   memcpy(pack.buff, pipe_pack.buff, pipe_pack.sizeVal-8);
   pack.sizeVal   = pipe_pack.sizeVal-8;
   memcpy(&pack.from_ip,   pipe_pack.buff+pipe_pack.sizeVal-8, sizeof(unsigned long));
   memcpy(&pack.from_port, pipe_pack.buff+pipe_pack.sizeVal-4, sizeof(unsigned long));

   unlock_pipe(INCOMING_PIPE);
   return;
}


/**************************************************************************/
/* this pipe should be readed only by one event processor, so   	  */
/* doesn't need lock&unlock it for reading. We should read 		  */
/* packet, ip, port, timeout, retry_num	and push them to inpack 	  */
/**************************************************************************/
void event_receive_packet(indirect_pack &inpack)
{
   pipe_pack.sizeVal = recvfrom(outgoing_pipe_fd, pipe_pack.buff, 
            MAX_PACKET_SIZE, 0, (struct sockaddr *)&op_serv_addr, 
	    &op_saddrlen);

   memcpy(inpack.buff, pipe_pack.buff, pipe_pack.sizeVal-8);
   inpack.sizeVal   = pipe_pack.sizeVal-24;
   memcpy(&inpack.to_ip,     pipe_pack.buff+pipe_pack.sizeVal-24, sizeof(unsigned long));
   memcpy(&inpack.to_port,   pipe_pack.buff+pipe_pack.sizeVal-20, sizeof(unsigned long));
   memcpy(&inpack.retry_num, pipe_pack.buff+pipe_pack.sizeVal-16, sizeof(unsigned long));
   memcpy(&inpack.time_out,  pipe_pack.buff+pipe_pack.sizeVal-12, sizeof(unsigned long));
   memcpy(&inpack.ask_stamp, pipe_pack.buff+pipe_pack.sizeVal-8,  sizeof(unsigned long));
   memcpy(&inpack.to_uin,    pipe_pack.buff+pipe_pack.sizeVal-4,  sizeof(unsigned long));

   return;
}

/**************************************************************************/
/* Write packet to outgoing pipe for event processor. We need		  */
/* addr:port, max number of retries (from config), max confirm		  */
/* timeout and long unique number of packet to recognize ask on 	  */
/* any protocol (ask_timestamp)						  */
/**************************************************************************/
void event_send_packet(Packet &pack, unsigned long to_uin, 
     unsigned long retry_num, unsigned long rtimeout, 
     unsigned long ask_timestamp)
{

   lock_pipe(OUTGOING_PIPE);

   /* preparing temp packet for pipe socket */
   memcpy(pipe_pack.buff, pack.buff, pack.sizeVal);
   pipe_pack.sizeVal = pack.sizeVal;
   pipe_pack.append();
   
   pipe_pack << (unsigned long)(ipToIcq(pack.from_ip))
             << (unsigned long)(pack.from_port)
	     << (unsigned long)retry_num
	     << (unsigned long)rtimeout
	     << (unsigned long)ask_timestamp
	     << (unsigned long)to_uin;
   
   /* send packet to pipe socket */
   sendto(outgoing_pipe_fd, pipe_pack.buff, pipe_pack.sizeVal, 0, 
          (struct sockaddr *)&op_serv_addr, op_saddrlen);

   unlock_pipe(OUTGOING_PIPE);
}

/**************************************************************************/
/* Incoming pipe initialization. It return socket descriptor 	 	  */
/**************************************************************************/
void incoming_pipe_init()
{
  if ((incoming_pipe_fd = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0)
  {
     LOG_SYS(0,("ERROR: Can't create unix domain socket for IPC pipe...\n"));
     exit(-1);
  }
  
  bzero(&ip_serv_addr, sizeof(ip_serv_addr));
  ip_serv_addr.sun_family = AF_UNIX;
  snprintf(ip_serv_addr.sun_path, sizeof(ip_serv_addr.sun_path)-1,
           "%s/pipe_inc.%u", lp_var_path(), lp_udp_port());
	   
  DEBUG(90, ("Binding pipe socket to: %s\n", ip_serv_addr.sun_path));
  
  ip_saddrlen = sizeof(ip_serv_addr.sun_family) + 
                strlen(ip_serv_addr.sun_path)+1;

  if (unlink(ip_serv_addr.sun_path) == 0)
  {
     LOG_SYS(0, ("Init: Unix socket (incoming) exist: unclean shutdown?...\n"));
  }

  if (bind(incoming_pipe_fd, (struct sockaddr *)&ip_serv_addr, ip_saddrlen) < 0)
  {
     LOG_SYS(0, ("ERROR: Can't bind pipe socket to special file...\n"));
     exit(-1);
  }
  
  chmod(ip_serv_addr.sun_path, 0600);
}

/**************************************************************************/
/* WWP pipe initialization. It return socket descriptor 	 	  */
/**************************************************************************/
void wwp_socket_init()
{
  if ((wwpsockfd = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0)
  {
     LOG_SYS(0,("ERROR: Can't create WWP unix domain socket...\n"));
     exit(-1);
  }
  
  bzero(&wp_serv_addr, sizeof(wp_serv_addr));
  wp_serv_addr.sun_family = AF_UNIX;
  snprintf(wp_serv_addr.sun_path, sizeof(wp_serv_addr.sun_path)-1,
           "/tmp/wwp_sock");
	   
  wp_saddrlen = sizeof(wp_serv_addr.sun_family) + 
                       strlen(wp_serv_addr.sun_path)+1;

  if (unlink(wp_serv_addr.sun_path) == 0)
  {
     LOG_SYS(0, ("Init: WWP Unix socket exist: unclean shutdown?...\n"));
  }

  if (bind(wwpsockfd, (struct sockaddr *)&wp_serv_addr, wp_saddrlen) < 0)
  {
     LOG_SYS(0, ("ERROR: Can't bind WWP socket to special file...\n"));
     exit(-1);
  }
  
  chmod(wp_serv_addr.sun_path, 0666);
}

/**************************************************************************/
/* Outgoing pipe initialisation. It return pipe descriptors pair 	  */
/**************************************************************************/
void outgoing_pipe_init()
{
  if ((outgoing_pipe_fd = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0)
  {
     LOG_SYS(0,("ERROR: Can't create unix domain socketfor IPC pipe...\n"));
     exit(-1);
  }
  
  bzero(&op_serv_addr, sizeof(op_serv_addr));
  op_serv_addr.sun_family = AF_UNIX;
  snprintf(op_serv_addr.sun_path, sizeof(op_serv_addr.sun_path)-1 ,
           "%s/pipe_out.%u", lp_var_path(), lp_udp_port());
	   
  DEBUG(90, ("Binding pipe socket to: %s\n", op_serv_addr.sun_path));
  
  op_saddrlen = sizeof(op_serv_addr.sun_family) + 
                strlen(op_serv_addr.sun_path)+1;

  if (unlink(op_serv_addr.sun_path) == 0)
  {
     LOG_SYS(0, ("Init: Unix socket (outgoing) exist: unclean shutdown?...\n"));
  }


  if (bind(outgoing_pipe_fd, (struct sockaddr *)&op_serv_addr, 
           op_saddrlen) < 0)
  {
     LOG_SYS(0, ("ERROR: Can't bind pipe socket to special file...\n"));
     exit(-1);
  }

  chmod(op_serv_addr.sun_path, 0600);
}

